#!/usr/bin/env python3
"""
Web Crawler Security
Advanced web application security crawler and analyzer.
Parameters:
    - url: Target URL to crawl
    - depth: Maximum crawl depth
    - output: Output file for results
Requires: requests, beautifulsoup4, selenium packages
"""
import argparse
import logging
from datetime import datetime
import json
import threading
import queue
from collections import defaultdict
import os
import re
import urllib.parse
import requests
from bs4 import BeautifulSoup
from selenium import webdriver
from selenium.webdriver.chrome.options import Options
from selenium.webdriver.common.by import By
from selenium.webdriver.support.ui import WebDriverWait
from selenium.webdriver.support import expected_conditions as EC
import concurrent.futures
import hashlib

class WebCrawlerSecurity:
    def __init__(self, url, depth=3, threads=5):
        self.base_url = url
        self.max_depth = depth
        self.num_threads = threads
        self.visited_urls = set()
        self.url_queue = queue.Queue()
        self.results = defaultdict(list)
        self.setup_logging()
        self.initialize_patterns()
        self.setup_browser()

    def setup_logging(self):
        logging.basicConfig(
            level=logging.INFO,
            format='%(asctime)s - %(levelname)s - %(message)s'
        )
        self.logger = logging.getLogger(__name__)

    def initialize_patterns(self):
        """Initialize security patterns to check"""
        self.patterns = {
            'sensitive_files': [
                r'\.git',
                r'\.env',
                r'\.config',
                r'\.log',
                r'\.bak',
                r'\.swp',
                r'\.old',
                r'\.backup',
                r'wp-config\.php',
                r'config\.php',
                r'database\.yml'
            ],
            'sensitive_data': [
                r'password',
                r'passwd',
                r'pwd',
                r'user',
                r'username',
                r'email',
                r'ssn',
                r'credit',
                r'card',
                r'token',
                r'api[_-]?key'
            ],
            'security_headers': [
                'X-Frame-Options',
                'X-XSS-Protection',
                'X-Content-Type-Options',
                'Strict-Transport-Security',
                'Content-Security-Policy',
                'X-Permitted-Cross-Domain-Policies',
                'Referrer-Policy',
                'Expect-CT',
                'Feature-Policy'
            ],
            'vulnerabilities': {
                'xss': [
                    r'<script>',
                    r'javascript:',
                    r'onerror=',
                    r'onload=',
                    r'eval\(',
                ],
                'sqli': [
                    r"'.*OR.*'1'.*='1",
                    r"admin'--",
                    r"UNION.*SELECT",
                    r"AND 1=1--",
                    r"OR 1=1--",
                ],
                'lfi': [
                    r'\.\./',
                    r'\.\.\\',
                    r'%2e%2e%2f',
                    r'%252e%252e%252f',
                ],
                'rce': [
                    r';.*\w+',
                    r'\|.*\w+',
                    r'`.*\w+`',
                    r'\$\(.*\)',
                ]
            }
        }

    def setup_browser(self):
        """Setup headless browser for JavaScript rendering"""
        try:
            chrome_options = Options()
            chrome_options.add_argument('--headless')
            chrome_options.add_argument('--no-sandbox')
            chrome_options.add_argument('--disable-dev-shm-usage')
            self.browser = webdriver.Chrome(options=chrome_options)
        except Exception as e:
            self.logger.error(f"Error setting up browser: {str(e)}")
            self.browser = None

    def is_same_domain(self, url):
        """Check if URL is from the same domain"""
        try:
            return urllib.parse.urlparse(url).netloc == (
                urllib.parse.urlparse(self.base_url).netloc
            )
        except:
            return False

    def normalize_url(self, url, base_url):
        """Normalize URL for consistency"""
        try:
            return urllib.parse.urljoin(base_url, url)
        except:
            return None

    def get_page_content(self, url):
        """Get page content with both requests and Selenium"""
        content = {
            'static': None,
            'dynamic': None,
            'headers': None
        }
        
        try:
            # Get static content
            response = requests.get(
                url,
                verify=False,
                timeout=10,
                headers={'User-Agent': 'SecurityCrawler/1.0'}
            )
            content['static'] = response.text
            content['headers'] = dict(response.headers)

            # Get dynamic content if browser is available
            if self.browser:
                self.browser.get(url)
                WebDriverWait(self.browser, 10).until(
                    EC.presence_of_element_located((By.TAG_NAME, "body"))
                )
                content['dynamic'] = self.browser.page_source
        except Exception as e:
            self.logger.error(f"Error getting content from {url}: {str(e)}")
        
        return content

    def extract_links(self, content, base_url):
        """Extract links from page content"""
        links = set()
        try:
            # Parse static content
            soup = BeautifulSoup(content['static'], 'html.parser')
            for tag in ['a', 'link', 'script', 'img', 'form']:
                for element in soup.find_all(tag):
                    if tag == 'a':
                        href = element.get('href')
                    elif tag in ['link', 'script', 'img']:
                        href = element.get('src')
                    elif tag == 'form':
                        href = element.get('action')
                    
                    if href:
                        normalized_url = self.normalize_url(href, base_url)
                        if normalized_url:
                            links.add(normalized_url)

            # Parse dynamic content if available
            if content['dynamic']:
                soup = BeautifulSoup(content['dynamic'], 'html.parser')
                for element in soup.find_all(['a', 'link', 'script']):
                    href = element.get('href') or element.get('src')
                    if href:
                        normalized_url = self.normalize_url(href, base_url)
                        if normalized_url:
                            links.add(normalized_url)
        except Exception as e:
            self.logger.error(f"Error extracting links: {str(e)}")
        
        return links

    def check_security_headers(self, headers):
        """Check for security headers"""
        missing_headers = []
        weak_headers = []
        
        for header in self.patterns['security_headers']:
            if header not in headers:
                missing_headers.append(header)
            else:
                value = headers[header]
                if header == 'X-Frame-Options':
                    if value.upper() not in ['DENY', 'SAMEORIGIN']:
                        weak_headers.append(header)
                elif header == 'X-XSS-Protection':
                    if value != '1; mode=block':
                        weak_headers.append(header)
                elif header == 'X-Content-Type-Options':
                    if value.lower() != 'nosniff':
                        weak_headers.append(header)
                elif header == 'Strict-Transport-Security':
                    if 'max-age' not in value.lower():
                        weak_headers.append(header)

        return {
            'missing': missing_headers,
            'weak': weak_headers
        }

    def check_sensitive_data(self, content):
        """Check for sensitive data exposure"""
        findings = []
        
        # Check in both static and dynamic content
        for content_type, html in [
            ('static', content['static']),
            ('dynamic', content['dynamic'])
        ]:
            if html:
                for pattern in self.patterns['sensitive_data']:
                    matches = re.finditer(
                        pattern,
                        html,
                        re.IGNORECASE
                    )
                    for match in matches:
                        context = html[
                            max(0, match.start() - 50):
                            min(len(html), match.end() + 50)
                        ]
                        findings.append({
                            'type': pattern,
                            'content_type': content_type,
                            'context': context.strip()
                        })

        return findings

    def check_vulnerabilities(self, url, content):
        """Check for common vulnerabilities"""
        vulnerabilities = defaultdict(list)
        
        # Check each vulnerability type
        for vuln_type, patterns in self.patterns['vulnerabilities'].items():
            for pattern in patterns:
                # Check in URL parameters
                if re.search(pattern, url, re.IGNORECASE):
                    vulnerabilities[vuln_type].append({
                        'location': 'url',
                        'evidence': url
                    })

                # Check in both static and dynamic content
                for content_type, html in [
                    ('static', content['static']),
                    ('dynamic', content['dynamic'])
                ]:
                    if html:
                        matches = re.finditer(
                            pattern,
                            html,
                            re.IGNORECASE
                        )
                        for match in matches:
                            context = html[
                                max(0, match.start() - 50):
                                min(len(html), match.end() + 50)
                            ]
                            vulnerabilities[vuln_type].append({
                                'location': f'{content_type}_content',
                                'evidence': context.strip()
                            })

        return dict(vulnerabilities)

    def check_forms(self, content):
        """Analyze HTML forms for security issues"""
        forms_info = []
        
        try:
            soup = BeautifulSoup(content['static'], 'html.parser')
            forms = soup.find_all('form')
            
            for form in forms:
                form_info = {
                    'action': form.get('action', ''),
                    'method': form.get('method', 'get').upper(),
                    'inputs': [],
                    'issues': []
                }

                # Check form method
                if form_info['method'] == 'GET':
                    form_info['issues'].append(
                        'Form uses GET method for submission'
                    )

                # Check for CSRF protection
                csrf_found = False
                for input_field in form.find_all('input'):
                    input_type = input_field.get('type', '')
                    input_name = input_field.get('name', '')
                    
                    if 'csrf' in input_name.lower():
                        csrf_found = True
                    
                    input_info = {
                        'type': input_type,
                        'name': input_name,
                        'issues': []
                    }

                    # Check for sensitive fields
                    if any(
                        pattern in input_name.lower()
                        for pattern in self.patterns['sensitive_data']
                    ):
                        if input_type != 'password':
                            input_info['issues'].append(
                                'Sensitive field not using password type'
                            )

                    form_info['inputs'].append(input_info)

                if not csrf_found:
                    form_info['issues'].append('No CSRF protection found')

                forms_info.append(form_info)

        except Exception as e:
            self.logger.error(f"Error analyzing forms: {str(e)}")

        return forms_info

    def crawl_url(self, url, depth):
        """Crawl single URL and check for security issues"""
        if depth > self.max_depth or url in self.visited_urls:
            return

        self.visited_urls.add(url)
        self.logger.info(f"Crawling {url} (depth: {depth})")

        try:
            # Get page content
            content = self.get_page_content(url)
            if not content['static']:
                return

            # Check security headers
            header_issues = self.check_security_headers(content['headers'])
            if header_issues['missing'] or header_issues['weak']:
                self.results['header_issues'].append({
                    'url': url,
                    'issues': header_issues
                })

            # Check for sensitive data
            sensitive_findings = self.check_sensitive_data(content)
            if sensitive_findings:
                self.results['sensitive_data'].append({
                    'url': url,
                    'findings': sensitive_findings
                })

            # Check for vulnerabilities
            vulns = self.check_vulnerabilities(url, content)
            if vulns:
                self.results['vulnerabilities'].append({
                    'url': url,
                    'findings': vulns
                })

            # Check forms
            forms = self.check_forms(content)
            if forms:
                self.results['forms'].append({
                    'url': url,
                    'forms': forms
                })

            # Extract and queue new links
            links = self.extract_links(content, url)
            for link in links:
                if (
                    self.is_same_domain(link) and
                    link not in self.visited_urls
                ):
                    self.url_queue.put((link, depth + 1))

        except Exception as e:
            self.logger.error(f"Error crawling {url}: {str(e)}")

    def crawler_worker(self):
        """Worker function for threaded crawling"""
        while True:
            try:
                url, depth = self.url_queue.get(timeout=1)
                self.crawl_url(url, depth)
                self.url_queue.task_done()
            except queue.Empty:
                break
            except Exception as e:
                self.logger.error(f"Worker error: {str(e)}")
                self.url_queue.task_done()

    def generate_report(self):
        """Generate security analysis report"""
        try:
            report = {
                'timestamp': datetime.now().isoformat(),
                'target': self.base_url,
                'scan_depth': self.max_depth,
                'urls_crawled': len(self.visited_urls),
                'findings': dict(self.results),
                'summary': {
                    'header_issues': len(self.results['header_issues']),
                    'sensitive_data': len(self.results['sensitive_data']),
                    'vulnerabilities': len(self.results['vulnerabilities']),
                    'form_issues': len(self.results['forms'])
                }
            }
            return report
        except Exception as e:
            self.logger.error(f"Error generating report: {str(e)}")
            return None

    def save_report(self, report, output_file):
        """Save analysis report to file"""
        try:
            with open(output_file, 'w') as f:
                json.dump(report, f, indent=4)
            self.logger.info(f"Report saved to {output_file}")
        except Exception as e:
            self.logger.error(f"Error saving report: {str(e)}")

    def print_summary(self, report):
        """Print analysis summary"""
        self.logger.info("\nWeb Security Analysis Summary:")
        self.logger.info(f"Target: {report['target']}")
        self.logger.info(f"Scan Depth: {report['scan_depth']}")
        self.logger.info(f"URLs Crawled: {report['urls_crawled']}")
        
        self.logger.info("\nFindings Summary:")
        for category, count in report['summary'].items():
            self.logger.info(f"{category.replace('_', ' ').title()}: {count}")

        # Print top issues
        if report['findings'].get('vulnerabilities'):
            self.logger.info("\nTop Vulnerabilities:")
            for vuln in report['findings']['vulnerabilities'][:5]:
                self.logger.info(f"\nURL: {vuln['url']}")
                for vuln_type, findings in vuln['findings'].items():
                    self.logger.info(f"{vuln_type}: {len(findings)} findings")

        if report['findings'].get('sensitive_data'):
            self.logger.info("\nSensitive Data Exposure:")
            for data in report['findings']['sensitive_data'][:5]:
                self.logger.info(f"\nURL: {data['url']}")
                self.logger.info(
                    f"Findings: {len(data['findings'])} instances"
                )

    def start_crawl(self, output_file=None):
        """Start security crawl analysis"""
        self.logger.info("Starting web security analysis")
        self.logger.info(f"Target: {self.base_url}")
        self.logger.info(f"Max Depth: {self.max_depth}")
        
        try:
            # Start with base URL
            self.url_queue.put((self.base_url, 0))
            
            # Create and start worker threads
            threads = []
            for _ in range(self.num_threads):
                t = threading.Thread(target=self.crawler_worker)
                t.start()
                threads.append(t)
            
            # Wait for all threads to complete
            for t in threads:
                t.join()
            
            # Generate and save report
            report = self.generate_report()
            if report and output_file:
                self.save_report(report, output_file)
            if report:
                self.print_summary(report)
                
        except KeyboardInterrupt:
            self.logger.info("Crawl interrupted by user")
        except Exception as e:
            self.logger.error(f"Error during crawl: {str(e)}")
        finally:
            if self.browser:
                self.browser.quit()

if __name__ == "__main__":
    parser = argparse.ArgumentParser(
        description="Web Security Crawler"
    )
    parser.add_argument(
        "url",
        help="Target URL to crawl"
    )
    parser.add_argument(
        "--depth",
        type=int,
        default=3,
        help="Maximum crawl depth"
    )
    parser.add_argument(
        "--threads",
        type=int,
        default=5,
        help="Number of crawler threads"
    )
    parser.add_argument(
        "--output",
        help="Output file for analysis report"
    )
    args = parser.parse_args()

    crawler = WebCrawlerSecurity(
        args.url,
        args.depth,
        args.threads
    )
    crawler.start_crawl(args.output)
